#!/usr/bin/python3
"""
Script to copy data from one Redis database to another and verify the copy operation.
The script prompts the user to enter the necessary credentials for both
the source and target Redis databases.

Requirements:
- See requirements.txt for pip packages
    use command: pip install -r /path/to/requirements.txt
- Redis server running and accessible
- Python 3.x

Usage:
1. Modify the connection constants within Configuration Section

- SOURCE_HOST
- SOURCE_PORT
- SOURCE_DB
- TARGET_HOST
- TARGET_PORT
- TARGET_DB

in the script to match your Redis server details.

2. Optionally, modify BATCH_SIZE and MAX_WORKERS to optimize
performance based on your system's capabilities.

3. Run the script using the command: python3 copy_redis_db.py

The script will prompt you to enter the following details:
- Source Username (optional, required for Redis v. 7 and higher)
- Source Password
- Target Username (optional, required for Redis v. 7 and higher)
- Target Password

Example:
Please enter the source Redis connection details:
Source Username (optional, required for Redis v. 7 and higher): source_username
Source Password: ********

Please enter the target Redis connection details:
Target Username (optional, required for Redis v. 7 and higher): target_username
Target Password: ********

The script will then copy the data from the source Redis database
to the target Redis database and verify the copy operation.
"""

import time
import getpass
import logging
import datetime
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from redis.exceptions import (
    RedisError,
    ConnectionError as RedisConnectionError,
    TimeoutError as RedisTimeoutError,
    ResponseError as RedisResponseError
)
import redis

# ===================================================================================
# Configuration Section starts here
# ===================================================================================
# This section contains configuration details for connecting to the Redis databases
# It's crucial to update these values according to your specific Redis setup
# Lines 72-75: Source Redis database connection details
# Lines 78-81: Target Redis database connection details
# ===================================================================================
# Source Redis database connection details
SOURCE_HOST = 'redxxxxx-master-1.data.eu01.onstackit.cloud'
SOURCE_PORT = 33333
SOURCE_DB = 0
SOURCE_SSL = True

# Target Redis database connection details
TARGET_HOST = 'redyyyyy-master-1.data.eu01.onstackit.cloud'
TARGET_PORT = 44444
TARGET_DB = 0
TARGET_SSL = True

# ===================================================================================
# Performance Tuning
# ===================================================================================
# Adjusting the batch size and number of threads can significantly impact
# the script's performance
# Line 95: Determines the number of keys to process in each batch
# Line 104: Specifies the maximum number of threads to use for parallel processing
# ===================================================================================
# BATCH_SIZE determines the number of keys to process in each batch.
# A larger batch size can improve performance but may use more memory.
# For most cases, a value between 500 and 2000 is reasonable.
# Adjust based on your system's memory capacity and performance needs.
BATCH_SIZE = 1000

# MAX_WORKERS specifies the maximum number of threads in the ThreadPoolExecutor
# for parallel processing.
# Increasing the number of threads can speed up the operation but may also
# increase CPU usage.
# Typically, setting this to the number of CPU cores or slightly less is a good
# starting point. Adjust based on your system's CPU capacity and workload.
# For systems with hyper-threading, consider starting with the number of logical processors.
MAX_WORKERS = 8
# ===================================================================================
# End of Configuration Section
# ===================================================================================


# Logger configuration
LOG_FORMAT = '%(asctime)s [%(levelname)-8s]: %(message)s'
time_now = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
logging.basicConfig(
    filename='redis_verification_' + time_now + '.log',
    encoding='utf-8',
    format=LOG_FORMAT,
    level=logging.INFO
)
logger = logging.getLogger(__name__)


def get_redis_credentials():
    """
    Prompt the user to enter Redis connection details interactively.
    
    :return: A dictionary containing the source and target Redis credentials.
    """

    print("Please enter the source Redis connection details:")
    source_username = input(
        "Source Database Username (optional, required for Redis v. 7 and higher): "
    )
    source_password = getpass.getpass("Source Database Password: ")

    print("\nPlease enter the target Redis connection details:")
    target_username = input(
        "Target Database Username (optional, required for Redis v. 7 and higher): "
    )
    target_password = getpass.getpass("Target Database Password: ")

    return {
        'source': {
            'username': source_username,
            'password': source_password
        },
        'target': {
            'username': target_username,
            'password': target_password
        }
    }


def create_redis_connection(host, port, db, password, username=None, ssl=True): # pylint: disable=too-many-arguments
    """
    Establishes an encrypted connection to a Redis database.
    
    :param host: Redis host address
    :param port: Redis port
    :param db: Database number
    :param password: Redis password
    :param username: Redis username (optional)
    :param ssl: Whether to use SSL for the connection (default: True)
    :return: Redis connection object
    """

    try:
        connection_params = {
            'host': host,
            'port': port,
            'db': db,
            'password': password,
            'ssl': ssl
        }
        if username:  # Check if username is provided
            connection_params['username'] = username

        connection = redis.Redis(**connection_params)
        # Simple connection test
        connection.ping()
        return connection

    except RedisConnectionError as conn_err:
        error_message = (
            f"Error: Could not establish connection to Redis at {host}:{port}.\n"
            f"Details: {conn_err}\n"
            "Possible causes:\n"
            "- Incorrect host, port, username, or password\n"
            "- Redis server is not running\n"
            "- Network issues\n"
            "- SSL settings mismatch\n"
            "Please check your connection settings and ensure the Redis server is accessible."
        )
        raise SystemExit(error_message) from conn_err

    except RedisResponseError as resp_err:
        error_message = (
            f"Authentication failed for Redis at {host}:{port}.\n"
            f"Details: {resp_err}\n"
            "Possible causes:\n"
            "- Incorrect username or password\n"
            "- User is disabled\n"
            "Please check your credentials."
        )
        raise SystemExit(error_message) from resp_err


def copy_batch(source_redis, target_redis, keys, start_index, end_index):
    """
    Copies a batch of keys from the source Redis instance to the target Redis instance.

    :param source_redis: Source Redis connection object
    :param target_redis: Target Redis connection object
    :param keys: List of keys to copy
    :param start_index: Starting index for the batch
    :param end_index: Ending index for the batch
    :return: True if all keys in the batch were successfully copied, False otherwise
    """

    pipe = target_redis.pipeline()

    success = True # Track overall success status

    for key in keys[start_index:end_index]:
        try:
            key_type = source_redis.type(key).decode('utf-8')
            if key_type == 'string':
                value = source_redis.get(key)
                ttl = source_redis.ttl(key)
                if value is not None:
                    pipe.set(key, value)
                    if ttl != -1:
                        pipe.expire(key, ttl)

            elif key_type == 'list':
                value = source_redis.lrange(key, 0, -1)
                ttl = source_redis.ttl(key)
                if value is not None:
                    pipe.rpush(key, *value)
                    if ttl != -1:
                        pipe.expire(key, ttl)

            elif key_type == 'set':
                value = source_redis.smembers(key)
                ttl = source_redis.ttl(key)
                if value is not None:
                    pipe.sadd(key, *value)
                    if ttl != -1:
                        pipe.expire(key, ttl)

            elif key_type == 'hash':
                value = source_redis.hgetall(key)
                ttl = source_redis.ttl(key)
                if value is not None:
                    pipe.hset(key, value)
                    if ttl != -1:
                        pipe.expire(key, ttl)

            elif key_type == 'zset':
                value = source_redis.zrange(key, 0, -1, withscores=True)
                ttl = source_redis.ttl(key)
                if value is not None:
                    pipe.zadd(key, dict(value))
                    if ttl != -1:
                        pipe.expire(key, ttl)

        except RedisError as e:
            #print(f"Error migrating key {key}: {e}")
            logger.error("Error migrating key %s: %s", key, e)
            success = False # Update overall success status

    pipe.execute()
    return success # Return overall success status


def copy_redis_db(source_host, source_port, source_db, source_password, source_ssl, # pylint: disable=too-many-arguments
                target_host, target_port, target_db, target_password, target_ssl,
                batch_size, max_workers,
                source_username=None, target_username=None):
    """
    Copies data from one Redis database to another, including handling
    of the username for Redis version 7 and above. 
    Uses threading to speed up the process.

    :param source_host: Source Redis host address
    :param source_port: Source Redis port
    :param source_db: Source database number
    :param source_password: Source Redis password
    :param source_ssl: Whether to use SSL for the source connection
    :param target_host: Target Redis host address
    :param target_port: Target Redis port
    :param target_db: Target database number
    :param target_password: Target Redis password
    :param target_ssl: Whether to use SSL for the target connection
    :param batch_size: Number of keys to copy in each batch (default: 1000)
    :param max_workers: Max Number of threads to use for ThreadPoolExecutor (default: 8)
    :param source_username: Source Redis username (optional)
    :param target_username: Target Redis username (optional)
    :return: Tuple (True if all data was successfully copied,
    False otherwise, total number of keys copied)
    """

    print(f"Connecting to source database at {source_host}:{source_port}, DB: {source_db}...")
    logger.info(
        "Connecting to source database at %s:%s, DB: %s...",
        source_host, source_port, source_db
    )

    source_redis = create_redis_connection(
        host=source_host,
        port=source_port,
        db=source_db,
        password=source_password,
        username=source_username,
        ssl=source_ssl
    )

    print(f"Connecting to target database at {target_host}:{target_port}, DB: {target_db}...")
    logger.info(
        "Connecting to target database at %s:%s, DB: %s...",
        target_host, target_port, target_db
    )

    target_redis = create_redis_connection(
        host=target_host,
        port=target_port,
        db=target_db,
        password=target_password,
        username=target_username,
        ssl=target_ssl
    )

    print("\nRetrieving keys from source database...")
    logger.info("Retrieving keys from source database...")

    # Retrieve all keys from source database
    # Using scan_iter for large databases
    # Iteratively retrieves keys from the Redis database in batches,
    # specified by the count parameter
    keys = [key.decode('utf-8') for key in source_redis.scan_iter(count=batch_size)]
    total_keys = len(keys)
    print(f"Found {total_keys} keys to copy.\n\nInitiating data copy process...")
    logger.info("Found %d keys to copy. Initiating data copy process...", total_keys)

    # Using ThreadPoolExecutor for thread management
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []  # List to hold Future objects returned by executor.submit()

        for i in range(0, total_keys, batch_size):
            start_index = i
            end_index = min(i + batch_size, total_keys)
            # Submit tasks to the executor and collect Future objects
            futures.append(
                executor.submit(
                    copy_batch, source_redis, target_redis, keys, start_index, end_index
                )
            )

        # Wait for all futures to complete and collect results
        results = [future.result() for future in as_completed(futures)]

    # Check results and log success or failure
    if not all(results):  # If any copy operation failed
        print("Data copy failed. Please check the log file for details.")
        logger.error("Data copy failed. Please check the log file for details.")

        return False, total_keys

    print("Data copy completed successfully.")
    logger.info("Data copy completed successfully.")

    return True, total_keys


def verify_batch(source_redis, target_redis, keys, start_index, end_index):
    """
    Verifies a batch of keys by comparing their values and TTLs between the source
    and target Redis databases.
    
    :param source_redis: Source Redis connection object
    :param target_redis: Target Redis connection object
    :param keys: List of keys to verify
    :param start_index: Starting index for the batch
    :param end_index: Ending index for the batch
    :return: True if batch verification was successfully, False otherwise
    """

    all_verified = True # Track overall verification status

    for key in keys[start_index:end_index]:

        try:

            key_type = source_redis.type(key).decode('utf-8')
            source_value = None
            target_value = None

            if key_type == 'string':
                source_value = source_redis.get(key)
                target_value = target_redis.get(key)

            elif key_type == 'list':
                source_value = source_redis.lrange(key, 0, -1)
                target_value = target_redis.lrange(key, 0, -1)

            elif key_type == 'set':
                source_value = source_redis.smembers(key)
                target_value = target_redis.smembers(key)

            elif key_type == 'hash':
                source_value = source_redis.hgetall(key)
                target_value = target_redis.hgetall(key)

            elif key_type == 'zset':
                source_value = source_redis.zrange(key, 0, -1, withscores=True)
                target_value = target_redis.zrange(key, 0, -1, withscores=True)

            if source_value != target_value:
                logger.error(
                    "Value for key %s does not match. Source: %s, Target: %s",
                    key, source_value, target_value
                )
                all_verified = False

            source_ttl = source_redis.ttl(key)
            target_ttl = target_redis.ttl(key)

            if source_ttl is None:
                source_ttl = -1

            if target_ttl is None:
                target_ttl = -1

            if source_ttl != target_ttl:
                logger.error(
                    "TTL for key %s does not match. Source TTL: %s, Target TTL: %s",
                    key, source_ttl, target_ttl
                )
                all_verified = False

        except (RedisError, RedisConnectionError, RedisTimeoutError) as e:
            logger.error("Error verifying key %s: %s", key, e)
            all_verified = False

    return all_verified # Return overall verification status


def verify_copy(source_host, source_port, source_db, source_password, source_ssl, # pylint: disable=too-many-arguments
                target_host, target_port, target_db, target_password, target_ssl,
                batch_size, max_workers,
                source_username=None, target_username=None):
    """
    Verifies the copy operation by comparing keys, values, and TTLs
    between the source and target Redis databases. 
    Uses threading for efficiency.

    :param source_host: Source Redis host address
    :param source_port: Source Redis port
    :param source_db: Source database number
    :param source_password: Source Redis password
    :param source_ssl: Whether to use SSL for the source connection
    :param target_host: Target Redis host address
    :param target_port: Target Redis port
    :param target_db: Target database number
    :param target_password: Target Redis password
    :param target_ssl: Whether to use SSL for the target connection
    :param batch_size: Number of keys to verify in each batch (default: 1000)
    :param max_workers: Max Number of threads to use for ThreadPoolExecutor (default: 8)
    :param source_username: Source Redis username (optional)
    :param target_username: Target Redis username (optional)
    :return: Tuple (True if all data was successfully copied and verified,
    False otherwise, total number of keys verified)
    """

    print(f"Connecting to source database at {source_host}:{source_port}, DB: {source_db}...")
    logger.info(
        "Connecting to source database at %s:%s, DB: %s...",
        source_host, source_port, source_db
    )

    source_redis = create_redis_connection(
        host=source_host,
        port=source_port,
        db=source_db,
        password=source_password,
        username=source_username,
        ssl=source_ssl
    )

    print(f"Connecting to target database at {target_host}:{target_port}, DB: {target_db}...")
    logger.info(
        "Connecting to target database at %s:%s, DB: %s...",
        target_host, target_port, target_db
    )

    target_redis = create_redis_connection(
        host=target_host,
        port=target_port,
        db=target_db,
        password=target_password,
        username=target_username,
        ssl=target_ssl
    )

    print("\nRetrieving keys from both databases...")
    logger.info("Retrieving keys from both databases...")
    source_keys = {key.decode('utf-8') for key in source_redis.scan_iter(count=batch_size)}
    target_keys = {key.decode('utf-8') for key in target_redis.scan_iter(count=batch_size)}

    # Compare key counts
    if len(source_keys) != len(target_keys):
        print("Key count does not match")
        logger.error("Keys do not match.")

        return False, len(source_keys)

    # Compare the keys
    if source_keys != target_keys:
        print("Keys do not match.")
        logger.error("Keys do not match.")

        return False, len(source_keys)

    # Convert source_keys to a list for indexing
    source_keys = list(source_keys)

    # Compare values and TTL for each key
    total_keys = len(source_keys)
    print(f"Total keys to verify: {total_keys}")
    logger.info("Total keys to verify: %d", total_keys)

    print("Verifying keys...")
    logger.info("Verifying keys...")

    # Compare values and TTL for each key using ThreadPoolExecutor
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []

        for i in range(0, len(source_keys), batch_size):
            start_index = i
            end_index = min(i + batch_size, len(source_keys))
            futures.append(
                executor.submit(
                    verify_batch, source_redis, target_redis, source_keys, start_index, end_index
                )
            )

        # Wait for all futures to complete and collect results
        results = [future.result() for future in as_completed(futures)]

    if not all(results):
        logger.error("Data verification failed. Please check the log file for details.")
        print("Data verification failed. Please check the log file for details.")

        return False, len(source_keys)

    print("\nAll data was successfully copied and verified.")
    logger.info("All data was successfully copied and verified.")

    return True, len(source_keys)


def main():
    """
    Main function that orchestrates the copy and verification processes,
    measuring and printing the time taken for each operation.
    """

    # Get credentials from the user input
    credentials = get_redis_credentials()

    # Source Redis database connection details
    source_username = credentials['source']['username']
    source_password= credentials['source']['password']

    # Target Redis database connection details
    target_username = credentials['target']['username']
    target_password = credentials['target']['password']

    # Copy data
    start_time = time.time()
    _, total_keys_copied = copy_redis_db(
        SOURCE_HOST, SOURCE_PORT, SOURCE_DB, source_password, SOURCE_SSL,
        TARGET_HOST, TARGET_PORT, TARGET_DB, target_password, TARGET_SSL,
        batch_size=BATCH_SIZE, max_workers=MAX_WORKERS,
        source_username=source_username, target_username=target_username
    )
    copy_duration = time.time() - start_time
    keys_per_second_copy = round(total_keys_copied / copy_duration)
    print(
        f"Copy operation finished in {copy_duration:.2f} seconds "
        f"({keys_per_second_copy} keys/s).\n"
    )
    logger.info(
        "Copy operation finished in %.2f seconds "
        "(%d keys/s).",
        copy_duration,
        keys_per_second_copy
    )

    # Verify copy
    start_time = time.time()
    _, total_keys_verified = verify_copy(
        SOURCE_HOST, SOURCE_PORT, SOURCE_DB, source_password, SOURCE_SSL,
        TARGET_HOST, TARGET_PORT, TARGET_DB, target_password, TARGET_SSL,
        batch_size=BATCH_SIZE, max_workers=MAX_WORKERS,
        source_username=source_username, target_username=target_username
    )
    verify_duration = time.time() - start_time
    keys_per_second_verify = round(total_keys_verified / verify_duration)
    print(
        f"Data verification completed in {verify_duration:.2f} seconds "
        f"({keys_per_second_verify} keys/s).\n"
    )
    logger.info(
        "Data verification completed in %.2f seconds "
        "(%d keys/s).",
        verify_duration,
        keys_per_second_verify
    )

if __name__ == "__main__":
    main()
